theory 
  Scenario1
imports 
  FWTesting
begin


subsubsection{* Network Model *}

text{* As there are only two distinct hosts in this model, we employ
the datatype address representation. The following definitions always
have to be provided when using the datatype representation, however
they are always standard (i.e. the same). The distribution already
provides the definitions for a datatype network consisting of three
elements. For the sake of completeness, we will not reuse that one
here.  *}

datatype hosts = Host1 | Host2

type_synonym HostNets = "hosts \<times> port \<times> protocol"

instance hosts :: adr .. 

defs (overloaded)
 src_port_def: "src_port (x::(HostNets,'b) packet) \<equiv> fst (snd (src x))"
 dest_port_def: "dest_port (x::(HostNets,'b) packet) \<equiv> fst (snd (dest x))"

defs (overloaded)
 src_protocol_def: "src_protocol (x::(HostNets,'b) packet) \<equiv> snd (snd (src x))"
 dest_protocol_def: "dest_protocol (x::(HostNets,'b) packet) \<equiv> snd (snd (dest x))"

definition
 host1 :: "HostNets net" where
"host1 = {{(a,b). a = Host1}}"

definition
 host2 :: "HostNets net" where
"host2 = {{(a,b). a = Host2}}"


subsubsection{* Policy Model *}

text{* We allow a couple of TCP and UDP ports in each direction and
deny everything else:*}

definition
 TestPolicy  where
"TestPolicy =  
             (allow_from_to_ports_prot tcp {(1024::port)..4096}  host1 host2 \<Oplus>
              allow_from_to_port_prot tcp (3389::port)           host1 host2 \<Oplus>
              allow_from_to_port_prot tcp (8080::port)           host1 host2 \<Oplus>
              allow_from_to_port_prot udp (8082::port)           host2 host2 \<Oplus>
              allow_from_to_port_prot tcp (21::port)             host2 host2 \<Oplus>
              allow_from_to_port_prot tcp (443::port)            host2 host2 \<Oplus>
              allow_from_to_port_prot udp (22::port)             host1 host2 \<Oplus>
              allow_from_to_port_prot tcp (23::port)             host1 host2 \<Oplus>       
              deny_all)"


subsubsection{* Simulation Experiments via the Code Generator. *}

text{* Isabelle/ISAR allows for access of the underyling SML environment, where functions
may be defined:*}
ML{*
fun fac 0 = 1
   |fac n = n * (fac (n -1)) 
*}
text{* ... and later on executed:*}
ML{* fac 10*}
text{* User-programmed functions may even refer to the internal Isabelle data-structures 
(the ML typesystem and Isabelle's Kernel Architecture will assure logical consistency) : *}
ML{* open Term;
     open Thm *}
text{* ... and logical inference operations:*}
ML{* sym RS sym*}

text{* So-called "antiquotations make it even possible to express links between
the model and the NL level*}
thm add_commute
ML{* @{thm add_commute} RS sym*}

text{* This may give you an idea how HOL-TestGen is internally built, and
why its implementation is actually "logically correct by construction".  *}

text{* A particular anti-quotation allows the user to invoke the code-generator
for (executable) logical HOL-expressions. Their values may be bound to variables
in the ML execution environment and used for computations: *}
ML{* val one = @{code "1::nat"}
     val suc = @{code "Suc"}
     val two = suc one*}

text{* Note that the current code generator version (Isabelle 2011) does not support
other input than constant symbols; synonyms, abbreviations etc. must be bound to a
constant symbol before:*}

definition "mt = Map.empty"  (* which is an abbreviation and is therefore no legal
                                argument in a code annotation. *)
definition "upd f a b = (f(a:= Some b))" (* the same *)

definition "lookup f x = f x"

definition "override f g = f \<Oplus> g"

text{* Now we can bind the entire infra-structure of partial maps ('data dictionaries') to
ML operators and execute them there. *}

ML{*

val mt = @{code mt}
fun upd f (x:''a) y = @{code upd}{equal = fn x => fn y => x=y} f x y
val dom = @{code dom}
val ran = @{code ran}
val lookup = @{code lookup}
val override = @{code "override"}
*}
text{*which allows us to do things like:*}
ML{* 
val A = upd(upd mt 1 1) 2 2
val B = upd mt 3 4
val Y = override A B
val Z = lookup Y 2
*}

subsubsection{* Definitions of the testing constraints. *}

text{* We model three testing constraints:
\begin{itemize} 
\item port\_positive: ensures that the destination port may be a 
positive numbers only
\item across\_hosts: ensures, that only packet sent
from one of the hosts to the respective other are considered
\item fix\_values: Fixes the source port, the source protocol and the content
part of each packet to a fixed (dummy) value
\end{itemize}
*}

definition port_positive :: "(HostNets,'b) packet \<Rightarrow> bool" where
 "port_positive x = (dest_port x > (0::port))"

definition
 accross_hosts :: "(HostNets,'b) packet \<Rightarrow> bool" where
"accross_hosts x = (((src x \<sqsubset> host1) \<and> (dest x \<sqsubset> host2)) \<or>  
                    ((src x \<sqsubset> host2)  \<and> (dest x \<sqsubset> host1)))" 

definition fix_values :: "(HostNets,DummyContent) packet \<Rightarrow> bool" where
 "fix_values x =  (src_port x = (1::port) \<and> src_protocol x = udp \<and> content x = data \<and> id x = 1)"


subsubsection{* Test Generation *} 

text {* Follwing the standard approach, all the lemmas which should be
applied to the test theorem are put into one "group lemma" *}

lemmas policyLemmas = TestPolicy_def host1_def fix_values_def port_positive_def 
                      host2_def accross_hosts_def packet_defs
                      src_port_def dest_port_def  
                      ProtocolPortCombinators.ProtocolCombinators 
                      src_protocol_def dest_protocol_def
                      
 text{* The test specification is standard: for all packets x for
 which the testing constraints hold, the firewall under test should
 exhibit the same behaviour as the modelled policy. *}

test_spec 
"port_positive x \<and> accross_hosts x \<and> fix_values x  \<longrightarrow> 
FUT x = TestPolicy x"
txt{* The following command puts the test theorem into the desired form. *}
apply  (prepare_fw_spec)

txt{* Next, the policy is unfolded and possibly simplified *}
apply (simp add: policyLemmas)

txt{* Test case generation, takes about half a minute in this example: *}
apply (gen_test_cases "FUT")

txt{* Simplification of the generated test cases. This makes test data generation more efficient. *}
apply (simp_all add: policyLemmas)

txt{*Store the test theorem *}
store_test_thm "Scenario1"


text{* Next, we set up the test data generator. First we enable the
external SMT solver. This if often required in the firewall
scenario. Please ensure to uncomment the line
\texttt{Z3\_NON\_COMMERCIAL="yes"} in the file
\texttt{/Isabelle20XX/contrib/z3-XXX/etc/settings} in order for
this to work. Next we define the numer of iterations. This is an
indication of how hard the random solver should try to find a valid
solution. This value needs to be increased if some test data cannot be
resolved. However, a fairly low value is usually fine when using the
SMT solver. *}

declare[[testgen_SMT=true]]
declare [[testgen_iterations=5]]


text {* We can now generate the test data. In this example this should
not take longer than a couple of seconds. The response should be
something like this:\begin{lstlisting} Test theorem (gen_test_data)
'Scenario1': 60 test cases in 4.107 seconds \end{lstlisting} *}

gen_test_data "Scenario1"


text {* The generated test data and the remaining proof obligations
(denoting unresolved test cases, i.e. there should best be none) can
be inspected by: *} 
thm Scenario1.test_data 
thm Scenario1.pos

text{* Finally, we can export the generated test data into a file by the following statement: *}

export_fw_data "Scenario1.data" Scenario1
end