(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: FTP_Test.thy 9257 2011-12-15 14:41:34Z lukasbru $ *)

header{*Test Data for FTP*}
theory FTP_Test
imports 
  FWTesting
begin


text{* In this theory we generate the test data for correct runs of
the FTP protocol. As usual, we start with definining the networks and
the policy. We use a rather simple policy which allows only FTP
connections starting from the Intranet and going to the Internet, and
deny everything else. *}

definition 
  intranet :: "IntegerPort net" where
  "intranet = {{(a,e) . a = 3}}"

definition 
  internet :: "IntegerPort net" where
  "internet = {{(a,c). a > 3}}"

definition 
  ftp_policy :: "(IntegerPort,ftp_msg) FWPolicy" where
  "ftp_policy = deny_all ++ allow_from_to_port (21::port) intranet  internet"

definition 
  ftp_policy1 :: "(IntegerPort,ftp_msg) FWPolicy" where
  "ftp_policy1 = allow_all"


text{* The next two constants check if an address is in the Intranet
or in the Internet respectively.*}

definition 
  is_in_intranet :: "IntegerPort  \<Rightarrow> bool" where
  "is_in_intranet a = ((fst a) = 3)"

definition 
  is_in_internet :: "IntegerPort \<Rightarrow> bool" where
  "is_in_internet a =  ((fst a) >3)"


text{*
The next definition is our starting state: an empty trace and the just defined policy.*}

definition 
  "\<sigma>_0_ftp" ::  "(IntegerPort, ftp_msg) history \<times>
                 (IntegerPort, ftp_msg) FWPolicy" 
where                                     
  "\<sigma>_0_ftp = ([],ftp_policy)"

text{*Next we state the conditions we have on our trace: a normal
behaviour FTP run from the intranet to some server in the internet on
port 21.*}

definition "accept_ftp" ::  "(IntegerPort, ftp_msg) history \<Rightarrow> bool" where
          "accept_ftp t = 
  (\<exists> c s i p. t \<in> NB_ftp c s i p \<and> is_in_intranet c \<and> is_in_internet s \<and> (snd s) = 21)"


text{*The depth of the test case generation corresponds to the maximal
length of generated traces. 4 is the minimum to get a full FTP
protocol run.  *}

declare [[testgen_depth=4]]


lemmas ST_simps = Let_def valid_SE_def unit_SE_def bind_SE_def
 in_subnet_def src_def dest_def IntegerPort.dest_port_def
 subnet_of_def id_def port_open_def is_init_def is_data_def
 is_port_request_def is_close_def p_accept_def content_def
 PolicyCombinators PortCombinators is_in_intranet_def
 is_in_internet_def intranet_def internet_def exI subnetOf_lemma
 subnetOf_lemma2 subnetOf_lemma3 subnetOf_lemma4 port_open_lemma
 ftp_policy_def allow_from_to_port_def NetworkCore.id_def
 IntegerPort.subnet_of_def


declare ST_simps[simp]

definition "ftp_p = ftp_policy"
 
lemma ftp_policy_charn: "(\<lambda>x. option_case (deny_all x) Some
              (if snd (fst (snd (snd x))) = 21
              then if (case fst (snd x) of (a, e) \<Rightarrow> a = 3) \<and> (case fst (snd (snd x)) of (a, c) \<Rightarrow> 3 < a)
                    then allow_all x else None
               else None)) = ftp_p"
apply (simp add: ST_simps ftp_p_def)
done

text {* Variant 1*}
text{* The test specification:*}
test_spec "\<lbrakk>ids1 (1::id) t; length t = length X; accept_ftp (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t FTP_ST; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept_ftp_def \<sigma>_0_ftp_def)
apply (rule impI)+
apply (unfold NB_ftp_def is_in_internet_def is_in_intranet_def)
apply simp
apply (gen_test_cases "FUT" split: HOL.split_if_asm)
apply (simp_all add: in_subnet_def subnet_of_def)
apply (simp_all add: ftp_policy_charn)
store_test_thm "ftp_test"

 
declare[[testgen_iterations=50]]
declare[[testgen_SMT=true]]

gen_test_data "ftp_test"

thm ftp_test.test_data
thm ftp_test.pos



text{* Much longer test sequences now *}

declare [[testgen_depth=7]]

(*
test_spec "\<lbrakk>ids1 (1::id) t; length t = length X; accept_ftp (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t FTP_ST; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept_ftp_def \<sigma>_0_ftp_def)
apply (rule impI)+
apply (unfold NB_ftp_def is_in_internet_def is_in_intranet_def)
apply simp
apply (gen_test_cases "FUT"  split: HOL.split_if_asm)
apply (simp_all add: id_def)
store_test_thm "ftp_test1a"

(*
Test theorem (gen_test_cases) 'ftp_test1a': 10 test cases in 220.942
seconds
*)

text{* We need to add all required lemmas to the simplifier set, such
that they can be used during test data generation. *}
declare[[testgen_iterations=10]]
declare[[testgen_SMT=true]]


gen_test_data "ftp_test1a"


thm ftp_test1a.test_data
thm ftp_test1a.pos

text {* With interleaving *}

declare [[testgen_depth=6]]
*)
definition "accept2_ftp" ::  "(IntegerPort, ftp_msg) history \<Rightarrow> bool" where
          "accept2_ftp t = 
  (\<exists> c1 s1 i1 p1 c2 s2 i2 p2. t \<in> ftp_2_interleaved c1 s1 i1 p1 c2 s2 i2 p2 \<and> 
                     is_in_intranet c1 \<and> is_in_internet s1 \<and> (snd s1) = 21 \<and>
                     is_in_intranet c2 \<and> is_in_internet s2 \<and> (snd s2) = 21 \<and>
                     i1 \<noteq> i2 \<and> p1 \<noteq> 21 \<and> p2 \<noteq> 21)"
(*

text{* The test specification:*}
test_spec "\<lbrakk>length t = length X; ids {1,2} t;
  accept2_ftp (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t  FTP_ST; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_other_def)
apply (rule impI)+
apply (unfold ftp_2_interleaved_def is_in_internet_def is_in_intranet_def)
apply simp
apply( simp split: option.splits decision.splits prod.splits)
apply (rotate_tac -1)
apply (thin_tac "?X")
apply simp
apply (rotate_tac -1)
apply (thin_tac "?X")
apply simp
apply (gen_test_cases "FUT" 6 1 split: HOL.split_if_asm simp: some1_equality) 
store_test_thm "ftp_test2"


(*Test theorem (gen_test_cases) 'ftp_test2': 40 test cases in 3120.833 seconds
*)


text{* We need to add all required lemmas to the simplifier set, such
that they can be used during test data generation. *}

declare[[testgen_iterations=10]]
declare[[testgen_SMT=true]]

gen_test_data "ftp_test2"



thm ftp_test2.test_data
thm ftp_test2.pos

export_fw_data "interleaved2.data" ftp_test2

*)
text{* Now one element longer*}
declare [[testgen_depth=7]]

fun is_ftp_packet where
  "is_ftp_packet (i,s,d,ftp_init) = True"
 |"is_ftp_packet (i,s,d,ftp_data) = True"
 |"is_ftp_packet (i,s,d,ftp_port_request x) = True"
 |"is_ftp_packet (i,s,d,ftp_close) = True"
 |"is_ftp_packet x = False"

fun is_ftp_trace where
 "is_ftp_trace (x#xs) = (is_ftp_packet x \<and> is_ftp_trace xs)"
|"is_ftp_trace [] = True"


declare [[testgen_depth=7]]


test_spec "\<lbrakk>length t = length X; ids {1,2} t; is_ftp_trace t;
  accept2_ftp (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t  FTP_ST; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept2_ftp_def \<sigma>_0_ftp_def is_other_def)
apply (rule impI)+
apply (unfold ftp_2_interleaved_def is_in_internet_def is_in_intranet_def)
apply simp
apply( simp split: option.splits decision.splits prod.splits)
apply (rotate_tac -1)
apply (thin_tac "?X")
apply simp
apply (rotate_tac -1)
apply (thin_tac "?X")
apply simp
apply safe
apply (gen_test_cases "FUT" 7 1 split: HOL.split_if_asm simp: some1_equality) 
store_test_thm "ftp_test2"


(*Test theorem (gen_test_cases) 'ftp_test2': 40 test cases in 3790.464 seconds*)

text{* We need to add all required lemmas to the simplifier set, such
that they can be used during test data generation. *}

declare[[testgen_iterations=10]]
declare[[testgen_SMT=true]]

gen_test_data "ftp_test2"



thm ftp_test2.test_data
thm ftp_test2.pos

export_fw_data "interleaved2.data" ftp_test2


declare ST_simps [simp del]


end
